/*
    Copyright 2011 Tianjun Liao Marco A. Montes de Oca, Dogan Aydin, Thomas Stuetzle and Marco Dorigo (IACOr-LS)
    Copyright 2010 Marco A. Montes de Oca, Dogan Aydin and Thomas Stuetzle (IPSOLS)
    Code structure is developed by Marco A. Montes de Oca.

    This file is part of IACOr-LS.

    IACOr-LS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    IACOr-LS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with IACOr-LS.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "configuration.h"
#include "rng.h"
#include "swarm.h"
#include "problem.h"
#include "ACO.h"
#include "localSearch.h"
#include "incrementalACO.h"

#include <iostream>
#include <cstdlib>
#include <cmath>

using namespace std;

IncrementalACO::IncrementalACO(Configuration* config, Swarm* swarm, Problem* problem) : ACO(config,swarm,problem)
{	
	ls = new LocalSearch(problem, config);
}

IncrementalACO::~IncrementalACO(){
	delete ls;
}

bool IncrementalACO::notDone(){

	if(p->getFunctionEvaluations() >= config->getMaxFES() || p->getBestSolutionValue() <= EPSILON)
		return false;
	return true;
}

void IncrementalACO::run(){

	unsigned int iterations=0;
	Particle* particle;
	const double* velocityVector;
	const double* positionVector;
	const double* bestPositionVector;
	const double* neighborBestPositionVector;
	double newVelocity[config->getProblemDimension()];
	double newPosition[config->getProblemDimension()];


	std::vector<int> neighbors;

	double stepsize;  
	double before,after;
	long double bestSoFar = LDBL_MAX;
	unsigned int noImprovement=0;
	do{

		for(int i=0;i<s->getSize();i++)
			(s->getParticle(i))->evaluateCurrentPosition();					

		if(config->getLocalSearchID() != NO_LOCAL_SEARCH){


			stepsize = RNG::randVal(0.5,0.5)*(config->getMaxInitRange()-config->getMinInitRange());

			particle = s->getBestParticle();
				if(s->getSize() == 1){
					//apply local search to the best particle
					before = particle->getBestValue();
					ls->applymtsls1(particle,stepsize);
					after = particle->getBestValue();
					//check the particle new position is better than the previous or not
					if(after < before)
						// if yes, reset failed attepmts
						particle->resetFailedAttempts();
					else 	// otherwise increment number of failed attempts
						particle->incrementFailedAttempts();
				}
				else{
					//Check the best particle is in local optima
					if(particle->getFailedAttempts()==config->getMaxRepeatedFailures()){	
						//if yes, apply local search to another (random) particle
						int pid = particle->getID();
						particle = s->getRandomParticle();
						while (particle->getID()==pid && particle->getFailedAttempts()>0)
							particle = s->getRandomParticle();	
					}

					// select a reference particle for determining step size adaptively
					Particle * reference = s->getRandomParticle();					
					while (reference->getID()==particle->getID())
						reference = s->getRandomParticle();
					
					// determine the max vector distance between the best(or random) and reference particle
					double max = 0.0;
					for(int j=0;j<config->getProblemDimension();j++){
						double temp = fabs(reference->getBestPosition()[j] - particle->getBestPosition()[j]);						
						if(temp > max)
							max = temp; 
					}
					// set step size as max vector distance 
					stepsize = max;

					if(stepsize <= EPSILON){
						stepsize = EPSILON;
					}

					before = particle->getBestValue();
					// apply local search
					ls->applymtsls1(particle,stepsize);
					after = particle->getBestValue();
					
					//check the particle new position is better than the previous or not
					if(after < before)
						particle->resetFailedAttempts();				
					else
						particle->incrementFailedAttempts();
				}
		}


		if(s->getSize()>1)

		{
		   acorselectway();
		}

		//Population growth and vertical social learning
		if(iterations>0 && iterations%config->getGrowthPeriod()==0 && s->getSize()<config->getMaximumPopulationSize()){
			s->addParticle(new Particle(config,p,s->getSize(),s->getBestParticle()));


		}

		//If improvement is lower than EPSILON (quite conservative)
	 	if(iterations > 0 && fabs(bestSoFar-p->getBestSolutionValue()) < EPSILON ){
                        noImprovement++;

                }
                else{
                        bestSoFar = p->getBestSolutionValue();
                        noImprovement=0;
                }

               if(noImprovement == config->getMaxStagnantIterations()){

//
			//destroy the current swarm 
			delete s;	
			//create a new swarm (Keeps best-so-far solution)
			s = new Swarm(config,p,p->getBestSoFarSolution(),p->getBestSolutionValue());
                       noImprovement=0;		
               }
		iterations++;
	}while(notDone());	
}

void IncrementalACO::acorexploit(){


	  double newacorPosition[config->getProblemDimension()];
		for(int n=0;n<config->getProblemDimension();n++)
		{
				//int l =randdiscrete();
				// get po value;
				double sums=0;// if int sums =0 divide function will be 0 " doulbe ()/double() " is right
				double sigma;

				for(int r=0;r<s->getSize();r++)
				{
				   sums=sums+double(abs(s->getParticle(r)->getBestPosition()[n]- s->getBestParticle()->getBestPosition()[n]))/(double (s->getSize())-double (1));
				}
				sigma=config->getxi()*sums;

				double randgaussian;


				    randgaussian=RNG::randGauss(sigma)+s->getBestParticle()->getBestPosition()[n];


					newacorPosition[n]=randgaussian;

		}
		s->getBestParticle()->setCurrentPosition(newacorPosition);


}
void IncrementalACO::acorexplore(){

	for(int m=0;m<s->getSize();m++)
	{

		    double newacorPosition[config->getProblemDimension()];
			for(int n=0;n<config->getProblemDimension();n++)
			{

					double sums=0;
					double sigma;

					for(int r=0;r<s->getSize();r++)
					{
					   sums=sums+double(abs(s->getParticle(r)->getBestPosition()[n]- s->getParticle(m)->getBestPosition()[n]))/(double (s->getSize())-double (1));
					}
					sigma=config->getxi()*sums;
					double randgaussian;


						randgaussian=RNG::randGauss(sigma)+s->getParticle(m)->getBestPosition()[n];


					    newacorPosition[n]=randgaussian;

			}


			s->getParticle(m)->setCurrentPosition(newacorPosition);



	}


}
void IncrementalACO::acorselectway(){
  double seletrand=RNG::randVal(0.0,1.0);
  if(config->getq()>seletrand)
  {
	  acorexplore();
  }
  else
  {
	  acorexploit();
  }


}

